from setuptools import setup

setup(
    name='property-caching',
    version='1.1.0',
    description='Property caching',
    author='Yola',
    author_email='engineers@yola.com',
    license='MIT (Expat)',
    url='https://github.com/yola/property-caching',
    packages=['property_caching'],
    test_suite='tests',
    classifiers=[
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
    ]
)
